#
# Copyright (C) 2016 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Unit tests for clicommand.py."""


import unittest

from bsp import device_stub
from cli import clicommand
from core import util_stub
from core import config_stub
from metrics import hit_store_stub
from metrics import metrics_util
from metrics.data_types import hit_stub
from project import target_stub


class ValidateTargetTest(unittest.TestCase):

    _BSP = 'devicename'

    def setUp(self):
        self.stub_util = util_stub.StubUtil()

        clicommand.util = self.stub_util

        self.dev = device_stub.StubDevice(self._BSP, downloaded=True)
        self.command = clicommand.Command()
        self.target = target_stub.StubTarget(
            name='test_target', os='brillo',
            os_version=self.stub_util.GetOSVersion(), board=self._BSP,
            board_version=self.dev.version, device=self.dev)

    def test_validate_valid(self):
        self.command.ValidateTarget(self.target)
        self.command.ValidateTarget(self.target, os=False)
        self.command.ValidateTarget(self.target, board=False)
        self.command.ValidateTarget(self.target, os=False, board=False)

    def test_validate_invalid_os(self):
        self.target.os = 'not_brillo'
        with self.assertRaises(clicommand.ValidationError):
            self.command.ValidateTarget(self.target)
        self.command.ValidateTarget(self.target, os=False)
        with self.assertRaises(clicommand.ValidationError):
            # Even if we ignore download state, the os still is invalid.
            self.command.ValidateTarget(self.target, os_downloaded=False)
        with self.assertRaises(clicommand.ValidationError):
            self.command.ValidateTarget(self.target, board=False)
        self.command.ValidateTarget(self.target, os=False, board=False)

    def test_validate_invalid_bsp(self):
        self.target.device_raises = True
        with self.assertRaises(clicommand.ValidationError):
            self.command.ValidateTarget(self.target)
        with self.assertRaises(clicommand.ValidationError):
            self.command.ValidateTarget(self.target, os=False)
        self.command.ValidateTarget(self.target, board=False)
        with self.assertRaises(clicommand.ValidationError):
            # Even if we ignore download state, the bsp still doesn't exist.
            self.command.ValidateTarget(self.target, board_downloaded=False)
        self.command.ValidateTarget(self.target, os=False, board=False)

    def test_validate_os_not_downloaded(self):
        self.target.os_version = 'not_a_version'
        with self.assertRaises(clicommand.ValidationError):
            self.command.ValidateTarget(self.target)
        self.command.ValidateTarget(self.target, os_downloaded=False)
        # Doesn't matter if os_downloaded is True as long as os is False
        self.command.ValidateTarget(self.target, os=False, os_downloaded=True)
        with self.assertRaises(clicommand.ValidationError):
            self.command.ValidateTarget(self.target, board=False)
        self.command.ValidateTarget(self.target, os=False, board=False)

    def test_validate_bsp_not_downloaded(self):
        self.dev.downloaded = False
        with self.assertRaises(clicommand.ValidationError):
            self.command.ValidateTarget(self.target)
        with self.assertRaises(clicommand.ValidationError):
            self.command.ValidateTarget(self.target, os=False)
        self.command.ValidateTarget(self.target, board_downloaded=False)
        # Doesn't matter if board_downloaded is True as long as board is False.
        self.command.ValidateTarget(self.target, board=False,
                                    board_downloaded=True)
        self.command.ValidateTarget(self.target, os=False, board=False)

    def test_validate_bsp_invalid_version(self):
        self.target.device_raises = True
        with self.assertRaises(clicommand.ValidationError):
            self.command.ValidateTarget(self.target)
        with self.assertRaises(clicommand.ValidationError):
            self.command.ValidateTarget(self.target, os=False)
        self.command.ValidateTarget(self.target, board=False)
        with self.assertRaises(clicommand.ValidationError):
            # Even if we ignore download state, the bsp still doesn't exist.
            self.command.ValidateTarget(self.target, board_downloaded=False)
        self.command.ValidateTarget(self.target, os=False, board=False)


def _keyboard_interrupt_func(_args):
    """Helper function to raise KeyboardInterrupt."""
    raise KeyboardInterrupt()


def _base_exception_func(_args):
    """Helper function to raise BaseException."""
    raise BaseException()


class RunWithMetricsTest(unittest.TestCase):
    """Tests to make sure RunWithMetrics() behaves as expected."""

    def setUp(self):
        self.config_stub = config_stub.StubConfig(metrics_opt_in='1')
        self.hit_stub = hit_stub.StubHit()
        self.hit_store_stub = hit_store_stub.StubHitStore()
        self.user_store_stub = self.config_stub.user_store

        clicommand.metrics_util.config = self.config_stub
        clicommand.metrics_util.hit = self.hit_stub
        clicommand.metrics_util.hit_store = self.hit_store_stub
        # pylint: disable=protected-access
        clicommand.metrics_util._MetricsUtilState.user_store = (
            self.user_store_stub)

        self.command = clicommand.Command()

    def last_uploaded_result(self):
        """Helper to return the most recent uploaded result or None."""
        if hit_stub.StubHit.last_send is None:
            return None
        return hit_stub.StubHit.get_custom_data_field(
            # pylint: disable=protected-access
            metrics_util._BRILLO_CD_RESULT)

    def test_upload_result(self):
        """Tests successfully uploading a Run() result."""
        for result in (0, 1, 130):
            self.command.Run = lambda _args, ret=result: ret

            self.assertEqual(result, self.command.RunWithMetrics([]))
            self.assertEqual(result, self.last_uploaded_result())

    def test_keyboard_interrupt(self):
        """Tests metrics are not uploaded after Ctrl+C."""
        self.command.Run = _keyboard_interrupt_func

        with self.assertRaises(KeyboardInterrupt):
            self.command.RunWithMetrics([])
        self.assertIsNone(self.last_uploaded_result())

    def test_base_exception(self):
        """Tests metrics are uploaded for any other exception."""
        self.command.Run = _base_exception_func

        with self.assertRaises(BaseException):
            self.command.RunWithMetrics([])
        self.assertEqual(1, self.last_uploaded_result())

    def test_metrics_opt_out(self):
        """Tests metrics are never sent if the user has opted out."""
        self.config_stub.user_store.metrics_opt_in = '0'

        self.command.Run = lambda _args: 0
        self.command.RunWithMetrics([])
        self.assertIsNone(self.last_uploaded_result())

        self.command.Run = lambda _args: 1
        self.command.RunWithMetrics([])
        self.assertIsNone(self.last_uploaded_result())

        self.command.Run = _keyboard_interrupt_func
        with self.assertRaises(KeyboardInterrupt):
            self.command.RunWithMetrics([])
        self.assertIsNone(self.last_uploaded_result())

        self.command.Run = _base_exception_func
        with self.assertRaises(BaseException):
            self.command.RunWithMetrics([])
        self.assertIsNone(self.last_uploaded_result())
