#
# Copyright (C) 2016 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Sets up a sysroot."""

from __future__ import print_function
import os
import shutil

from bsp import device
from cli import clicommand
from core import util
from environment import sysroot
from environment import sysroot_util
from environment import toolchain_util


# Note: this should be kept in sync with project.target.PLATFORM_CACHE_FORMAT
SAMPLE_MAKEFILE_TEMPLATE = """\
BDK={bdk_command}
DEVICE={board}.{board_version}
OS={os}.{os_version}

PLATFORM_CACHE:=$(shell $(BDK) config check platform_cache)/$(OS)/$(DEVICE)
SYSROOT=$(PLATFORM_CACHE)/sysroot
TOOLCHAIN=$(PLATFORM_CACHE)/toolchain
PKG_CONFIG_VARS=PKG_CONFIG_LIBDIR="$(SYSROOT)/usr/share/pkgconfig"\
 PKG_CONFIG_SYSROOT_DIR="$(SYSROOT)"

CXX=$(TOOLCHAIN)/g++
DEPS=
ifdef DEPS
CFLAGS=`$(PKG_CONFIG_VARS) pkg-config --cflags $(DEPS)`
LDFLAGS=`$(PKG_CONFIG_VARS) pkg-config --libs $(DEPS)`
endif

all: myproject

myproject: main.cpp
\t$(CXX) $(CFLAGS) $(LDFLAGS) --sysroot=$(SYSROOT) main.cpp -o myproject

clean:
\trm myproject
"""


class Setup(clicommand.Command):
    """Set up a sysroot in the given directory."""

    @staticmethod
    def Args(parser):
        Setup.AddProductArgs(parser)
        # TODO(b/28249627): This should be removed and replaced by some form
        #     of config.
        parser.add_argument('--pkgconfig_csv',
                            help='The path to a csv file describing packages '
                            'for the sysroot (default {}).'.format(
                                util.GetBDKPath('pkgconfig', 'packages.csv')),
                            default=util.GetBDKPath('pkgconfig',
                                                    'packages.csv'))

    def Run(self, args):
        """Sets up a sysroot and toolchain."""
        try:
            _, target = self.GetSpecAndTarget(args)
        except clicommand.Error as error:
            print(error)
            return 1
        device_ = target.get_device()

        # TODO(b/27385334): maybe toolchain/sysroot setup should really be
        #     different commands, with a helper command that invokes both.
        #     Alternatively, they could work together better such as by adding
        #     the necessary flags to use the sysroot into the tools.

        # Check args.
        platform_build_cache = target.platform_build_cache()
        product_out = util.GetAndroidProductOut(platform_build_cache,
                                                target.board)
        if not os.path.isdir(product_out):
            print('Could not find platform build output "{}". '
                  'You must run\n'
                  '    bdk build platform\n'
                  'before you set up your build environment.'.format(
                      product_out))
            return 1
        if device_.arch not in sysroot_util.SysrootUtil.DEFAULT_ARCH_INCLUDES:
            print('Unknown architecture: {}'.format(device_.arch))
            return 1
        if not os.path.isfile(args.pkgconfig_csv):
            print('No such file: {}'.format(args.pkgconfig_csv))
            return 1
        try:
            host_arch = util.GetHostArch()
        except util.HostUnsupportedArchError as e:
            print('Could not set up environment: {}'.format(e))
            return 1

        # Basic values.
        shared_libs = os.path.join(product_out, 'obj', 'lib')
        static_libs = os.path.join(product_out, 'obj', 'STATIC_LIBRARIES')

        # TODO(b/27458309): determine an actual location for this.
        board_packages = util.GetOSPath(target.os_version,
                                        'device', device_.vendor, target.board,
                                        'developer', 'packages.csv')
        sysroot_path = target.platform_cache('sysroot')
        toolchain_path = target.platform_cache('toolchain')

        ret = 0
        # Generate the sysroot.
        if os.path.exists(sysroot_path):
            print ('Not generating sysroot: '
                   'something already exists at {}.'.format(sysroot_path))
        else:
            sysroot_success = False
            brillo_sysroot = sysroot.Sysroot(sysroot_path)
            setup_util = sysroot_util.SysrootUtil(brillo_sysroot,
                                                  target.os_version)
            # Setup from a base, add libs.
            try:
                setup_util.SetupBaseSysroot(device_.arch, shared_libs,
                                            static_libs)
                setup_util.AddSharedLibsFromCSV(shared_libs, args.pkgconfig_csv)
                # Add board libs, if any
                with device_.linked(target.os_version):
                    if os.path.isfile(board_packages):
                        setup_util.AddSharedLibsFromCSV(shared_libs,
                                                        board_packages,
                                                        suffix=target.board)
                sysroot_success = True
            except (IOError, ValueError, device.Error) as e:
                print('Failed to generate sysroot: {}'.format(e))
            finally:
                if not sysroot_success:
                    ret = 1
                    brillo_sysroot.Destroy()

        # Generate toolchain.
        if os.path.exists(toolchain_path):
            print('Not generating toolchain: '
                  'something already exists at {}.'.format(toolchain_path))
        else:
            toolchain_success = False
            try:
                os.makedirs(toolchain_path)
                toolchain_util.GenerateToolchain(target.os_version, host_arch,
                                                 device_.arch, toolchain_path)
                toolchain_success = True
            except toolchain_util.Error as e:
                print('Failed to generate toolchain: {}'.format(e))
            finally:
                if not toolchain_success:
                    ret = 1
                    shutil.rmtree(toolchain_path)

        if not ret:
            print('Finished building environment.')
            print('Sample Makefile:\n\n' + SAMPLE_MAKEFILE_TEMPLATE.format(
                bdk_command=util.GetBDKPath('cli/bdk'), os=target.os,
                os_version=target.os_version, board=target.board,
                board_version=target.board_version))

        return ret
