#
# Copyright (C) 2016 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Functions for building an image."""


import fileinput
import os
import shutil
import subprocess

from core import util
from environment import sysroot
import error
from project import pack
from project import packs
from selinux import policy


# Allowed image types for BuildImage.
IMAGE_TYPE_ODM = 'odm'
IMAGE_TYPE_SYSTEM = 'system'
IMAGE_TYPES = [IMAGE_TYPE_ODM, IMAGE_TYPE_SYSTEM]


class Error(error.Error):
    """General build failure."""


class PathError(Error):
    """Raised when a provided path does not meet expectations."""


class ImageTypeError(Error):
    """Raise when an unknown image type is seen."""


def _CheckBuildImagePaths(cache_dir, output_dir, product_out, host_tools,
                          build_tools):
    """Checks that all paths necessary for an image build meet expectations.

    Does not validate the contents of any paths.
    Will create an empty output_dir if it doesn't already exist.

    Args:
        cache_dir: Directory where user files and system files are merged.
            (This is normally per-target, not the global cache dir.)
        output_dir: Directory where image build output should be placed.
        product_out: Directory where the built platform can be found.
        host_tools: Directory where host tools can be found.
        build_tools: Directory where build tools can be found.

    Raises:
        PathError: An expected directory cannot be found or isn't a dir.
    """
    if not os.path.isdir(cache_dir):
        raise PathError('Cache dir "{}" is not a directory.'.format(cache_dir))
    if os.path.isfile(output_dir):
        raise PathError('Can not create output dir "{}", is a file'.format(
            output_dir))
    if not os.path.isdir(product_out):
        raise PathError(
            'Could not find platform build output "{}". '
            'Use `bdk build platform` to build it.'.format(
                product_out))
    if not os.path.isdir(host_tools):
        # Probably tampered with/not built.
        raise PathError(
            'Could not find host tools directory ({}). '
            'Use `bdk build platform` to build it.'.format(host_tools))
    if not os.path.isdir(build_tools):
        # This should only be missing if it was tampered with.
        raise PathError(
            'Could not find build tools directory ({}). '
            'You may need to re-download the OS version.'.format(
                build_tools))

    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)


def AddTargetOsPacks(spec, target):
    """Adds target OS packs to the spec.

    Args:
        spec: project.project_spec.ProjectSpec to add the OS packs to.
        target: project.target.Target object to get OS info from.
    """
    platform_dir = target.platform_build_cache()
    product_out = util.GetAndroidProductOut(platform_dir, target.board)
    os_packs = packs.Packs()
    os_packs.namespace = target.os_namespace
    os_core = pack.Pack(os_packs.namespace, 'generated_system')
    os_core.add_provides('os.core')
    base_copy = pack.Copy(os_core)
    base_copy.src = os.path.join(product_out, 'system')
    base_copy.src_type = pack.CopyType.DIR
    base_copy.recurse = True
    base_copy.dst = '/system'
    base_copy.dst_type = pack.CopyType.DIR
    # Do not generate fs_config_* or file_context_* for this pack.
    base_copy.acl.override_build = False
    os_core.add_copy(base_copy)
    os_packs.add_pack(os_core)
    spec.add_packs(os_packs)


def CreateTargetCache(spec, target, cache_dir, mountpoint='/', update=True,
                      verbose=True):
    """Copies all files specified in the target to a single cache dir.

    This call will create a submap from the parent spec and then use the
    resulting copy destinations to populate the supplied path as if it were
    the root of the destination.

    In the future, non-os and non-board packs will likely end up in their
    own image (not /system), so this will create that cache ignoring any
    Copy() nodes owned by a Pack that is prefixed by the target os or board.

    Args:
        spec: project.ProjectSpec global specification
        target: project.target.Target to cache
        cache_dir: path to place the root/ of copied files.
        mountpoint: optional path prefix to exclusively cache.
        update: optionally, only replaces files that have changed.
        verbose: (optional). If True, print information about what's
            happening. Default True.

    Returns:
        dict of uncached { dst_path => copy }

    Raises:
        project.dependency.Error: If there is an unfulfilled dependency.
        PathError: If a required path is missing.
    """
    target_map = target.create_submap(spec.packmap)

    cache_root = os.path.join(cache_dir, 'root')
    cache = sysroot.Sysroot(cache_root, copy_newer_only=update)
    fs_config_files = {}
    fs_config_dirs = {}
    file_context = set()
    uncached = {}

    # pack.Copy._reconcile_paths() *should* make these always what is seen here.
    # TODO(wad): Add a <link> node.
    allowed_copies = [
        (pack.CopyType.FILE, pack.CopyType.FILE, False),
        (pack.CopyType.GLOB, pack.CopyType.DIR, False),
        (pack.CopyType.GLOB, pack.CopyType.DIR, True),
        (pack.CopyType.DIR, pack.CopyType.DIR, False),
        (pack.CopyType.DIR, pack.CopyType.DIR, True),
    ]

    final_files = {}
    for destination, copy in target_map.copy_destinations.iteritems():
        copy = copy[0]  # deduping has already occurred.

        # Currently, skip unknown mountpoints.
        if not destination.startswith(mountpoint):
            uncached[destination] = copy
            continue

        # All destinations are required to be absolute.
        # Make them relative so they play nicely with Sysroot().
        destination = destination.lstrip('/')

        # Proactively check allowed combinations so that the error
        # handling isn't interleaved with logic.
        if (copy.src_type, copy.dst_type, copy.recurse) not in allowed_copies:
            # TODO(wad): Should these be FileABugError()s?
            raise PathError(
                '{}: impossible copy reached: {}'.format(copy.pack.origin,
                                                         copy))

        # TODO(arihc): Will os packs ever draw in files from BSPs
        #     (other than build output files)? If so, need to link BSP to OS.
        files = set()
        try:
            if copy.src_type == pack.CopyType.FILE:  # FILE -> FILE
                cache.Makedirs(os.path.dirname(destination))
                if os.path.islink(copy.src):
                    # TODO(wad): Resolve how to skip permission setting on
                    #     symlinks.
                    cache.AddSymlink(copy.src, destination)
                else:
                    cache.AddFile(copy.src, destination)
                files.add(destination)
            elif copy.src_type == pack.CopyType.GLOB:  # GLOB -> DIR
                files.update(cache.AddGlob(copy.src, destination,
                                           recurse=copy.recurse))
            elif copy.src_type == pack.CopyType.DIR:  # DIR -> DIR
                files.update(cache.AddDir(copy.src, destination,
                                          recurse=copy.recurse, symlinks=True))
                fs_config_dirs[destination] = copy.acl.fs_config(binary=True)
            for f in files:
                # Even though the fs_config_files format support globs, the
                # final list of files is known at this point.
                fs_config_files[f] = copy.acl.fs_config(path=f, binary=True)
                if copy.acl.selabel:
                    file_context.update([copy.acl.file_context(path=f)])
                final_files[os.path.sep + f] = copy

        except IOError as e:
            # Annotate the error with the line that where the <copy> is defined.
            raise PathError('{}: {}'.format(copy.origin, e))

    # Walk the cache tree and remove any unmanaged dirs. First
    # we compute all allowed subpaths from the list of copied files, then
    # walk the cache tree and remove any unknown files and directories.
    final_dirs = set('/')
    for dst, copy in final_files.iteritems():
        # Final files is always a full path so let's drop the file.
        dst = os.path.dirname(dst)
        while dst != os.path.sep:
            # TODO(b/27848879) Add <dir> node set-acl support.
            tmp_acl = pack.Copy(None, dst).acl
            tmp_acl.perms = '0555'  # Default to rx for dirs.
            tmp_acl.override_build = copy.acl.override_build
            if dst not in fs_config_dirs:
                fs_config_dirs[dst] = '{}'.format(
                    tmp_acl.fs_config(binary=True))
            # The default root map covers file_context for now.
            final_dirs.add(dst)
            dst = os.path.dirname(dst)

    all_paths = set()
    all_files = set()
    # Walk the tree as quickly as possible with os.walk.
    # Use set differencing to compute the unmanaged entries.
    for root, _, files in os.walk(cache_root):
        dst_path = root[len(cache_root):] or os.path.sep
        # Don't touch unmanaged files outside of the given mountpoint.
        if not dst_path.startswith(mountpoint):
            continue
        all_paths.add(dst_path)
        all_files.update(set([os.path.join(dst_path, f) for f in files]))
    unmanaged_file = all_files.difference(final_files.keys())
    unmanaged_path = all_paths.difference(final_dirs)

    if verbose and (len(unmanaged_file) or len(unmanaged_path)):
        print 'Removing unmanaged paths:\n {}'.format(
            '\n '.join(sorted(unmanaged_file.union(unmanaged_path))))
    for f in unmanaged_file:
        os.remove(os.path.join(cache_root, f.lstrip(os.path.sep)))
    for p in reversed(sorted(unmanaged_path)):
        os.rmdir(os.path.join(cache_root, p.lstrip(os.path.sep)))

    # Sort the list from most specific to least specific.
    dsts = sorted(fs_config_dirs.keys(), key=lambda k: k.count(os.path.sep),
                  reverse=True)
    sorted_dirs = list()
    for dst in dsts:
        sorted_dirs.append(fs_config_dirs[dst])

    # Create fs_config_{files, dirs} and file_context build metadata
    # files outside of the sysroot to avoid project spec conflicts.
    if not os.path.isdir(os.path.join(cache_dir, 'etc')):
        os.makedirs(os.path.join(cache_dir, 'etc'))
    # List of (base file name, generated content) to make code sharing easier.
    metadata_files = [
        (('etc', 'fs_config_dirs'), sorted_dirs),
        (('etc', 'fs_config_files'), fs_config_files.values()),
    ]
    # Walk the base names, create the file, then append any pre-existing content.
    for name, data in metadata_files:
        data_file = os.path.join(cache_dir, *name)
        tgt_file = os.path.join(cache_root, 'system', *name)
        with open(data_file, 'w') as f:
            f.write(''.join(data))
            if os.path.isfile(tgt_file):
                with open(tgt_file, 'r') as extra:
                    shutil.copyfileobj(extra, f)

    with open(os.path.join(cache_dir, 'file_contexts'), 'w') as f:
        f.write('\n'.join(file_context))

    return uncached


def _CreateBuildProps(build_root, product_out, image_type, info_file):
    if image_type == IMAGE_TYPE_ODM:
        build_props = {
            'mount_point': 'odm',
            'fs_type': 'ext4',
            # TODO(b/27854052): Get the size in here.
            'partition_size': '134217728',
            'extfs_sparse_flag': '-s',
            'skip_fsck': 'true',
            'selinux_fc': build_root.Path('root', 'file_contexts.bin'),
        }
        build_root.WriteFile(info_file, '\n'.join(
            ['{}={}'.format(k, v) for k, v in build_props.iteritems()]))
    elif image_type == IMAGE_TYPE_SYSTEM:
        # Add and update a copy of system_image_info.txt
        build_root.AddFile(os.path.join(product_out,
                                        'obj',
                                        'PACKAGING',
                                        'systemimage_intermediates',
                                        'system_image_info.txt'), info_file)
        for line in fileinput.input(build_root.Path(info_file), inplace=True):
            # We just want to change the SELinux file contexts.
            if line.startswith('selinux_fc'):
                line = 'selinux_fc=' + build_root.Path('root',
                                                       'file_contexts.bin')
            print line.rstrip()
    else:
        build_root.WriteFile(info_file, 'mount_point={}'.format(image_type))


def BuildImage(image_type, target, cache_dir, output_dir):
    """Builds the image specified by image_type.

    Caller should validate target OS before calling.

    Args:
        image_type: The type of the image. One of IMAGE_TYPES.
        target: project.target.Target to build an image of.
        cache_dir: Directory to build the image in. This path
             may be prepared by prior callers. E.g., CreateTargetCache().
        output_dir: Directory to place built .img file in.

    Returns:
        The image build exit code.

    Raises:
        PathError: An expected directory cannot be found or isn't a dir.
        util.HostUnsupportedArchError: The host is an unsupported architecture.
        ImageTypeError: An invalid parameter value has been supplied for
            image_type.
    """
    if image_type not in IMAGE_TYPES:
        raise ImageTypeError('image_type must be one of {}: {}'.format(
            IMAGE_TYPES, image_type))
    # Set some useful variables.
    build_tools = util.GetOSPath(target.os_version,
                                 'build', 'tools', 'releasetools')
    platform_out = target.platform_build_cache()
    product_out = util.GetAndroidProductOut(platform_out, target.board)
    host_arch = util.GetHostArch()
    host_tools = os.path.join(platform_out, 'host', host_arch, 'bin')
    output_file = os.path.join(output_dir, '{}.img'.format(image_type))
    ret = 1

    # Check that all required inputs and tools exist.
    _CheckBuildImagePaths(cache_dir, output_dir, product_out, host_tools,
                          build_tools)
    # Check to ensure there are entries in root/image-mountpoint.
    if not os.path.exists(os.path.join(cache_dir, 'root', image_type)):
        print ('Warning: The specification does not define any destinations '
               'for this image type: {}'.format(image_type))

    build_root = sysroot.Sysroot(cache_dir, copy_newer_only=True)

    # Build 'filecontexts.bin' and 'sepolicy' SELinux files.
    with target.get_device().linked(target.os_version):
        policy.BuildSepolicy(target, platform_out, cache_dir)
        policy.BuildFileContexts(target, platform_out, cache_dir)

    _CreateBuildProps(build_root, product_out, image_type, 'image_info.txt')

    # Build an image from the build root.
    additional_path = host_tools + os.pathsep + build_tools
    if 'PATH' in os.environ:
        os.environ['PATH'] += os.pathsep + additional_path
    else:
        os.environ['PATH'] = additional_path

    # Repoint product out so that changes made locally are reflected
    # for the build tooling.    This primarily enables the discovery
    # of etc/fs_config_{files, dirs}.
    product_out = build_root.Path()
    ret = subprocess.call(['build_image.py',
                           build_root.Path('root', image_type),
                           build_root.Path('image_info.txt'),
                           output_file,
                           product_out])
    return ret
