#
# Copyright (C) 2015 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Product Templates

   This file contains the simplest possible templates for generating
   scaffolding product files.
"""

import string

ANDROIDPRODUCTS_MK = string.Template("""\
# Automatically generated by bdk.
# When editing, always keep the primary product makefile first.
PRODUCT_MAKEFILES := $$(LOCAL_DIR)/${name}.mk
""")

PRODUCT_MK = string.Template("""\
# Automatically generated by bdk.
# Do not edit. Use extras.mk instead.
CFGTREE_ROOT := $$(LOCAL_PATH)
include device/generic/brillo/brillo_base.mk

PRODUCT_NAME := $$(call get_product_name_from_file)
ifneq ($$(PRODUCT_NAME),$$(call cfgtree-get,name))
  $$(error config/name, <name>.mk, and AndroidProducts.mk must all be updated together.)
endif
PRODUCT_BRAND := $$(call cfgtree-get,brand)
PRODUCT_DEVICE := $$(call cfgtree-get,device)
PRODUCT_MANUFACTURER := $$(call cfgtree-get,manufacturer)
PRODUCT_PACKAGES += $$(call cfgtree-get-if-exists,packages)
# Populate PRODUCT_COPY_FILES from disk but avoid requiring the use of
# $$(LOCAL_PATH) by prepending it here.
PRODUCT_COPY_FILES += $$(addprefix $$(LOCAL_PATH)/,$$(call cfgtree-get-if-exists,copy_files))

# Expect any required SELinux policies in the sepolicy/ path.
BOARD_SEPOLICY_DIRS := $$(BOARD_SEPOLICY_DIRS) $$(LOCAL_PATH)/sepolicy

# Include any configured 3rd party packages.
PRODUCT_3RD_PARTY_PACKAGES += $$(call cfgtree-get-if-exists,3rd-party-packages)

# Add any extra Android product.mk directive in extras.mk
-include $$(LOCAL_PATH)/extras.mk
""")

BDK_BANNER = string.Template("""\
bdk_banner() {
  echo Brillo Development Kit v."${bdk_version}" from "${bdk_path}"
  echo
  echo "${bdk_warning}"
  echo
  echo Available commands:
  echo   bdk, m, mm, adb, fastboot, provision, gdbclient.py, brunch
}
bdk_banner

""")

ENV_EXPORTS = string.Template("""\
export BDK_PATH="${bdk_path}"
# Put the bdk executable in your path.
export PATH=${cli_path}:$$PATH
""")

ENV_ALIASES = string.Template("""\

brunch() {
  bdk "$$@"
}

m() {
  bdk product build -- "$$@"
}

mm() {
  bdk product build -s -- "$$@"
}

adb() {
  bdk product tool adb "$$@"
}

fastboot() {
  bdk product tool fastboot "$$@"
}

provision() {
  bdk product provision "$$@"
}

# Naming a shell function as a Python script is pretty ugly, but:
#  1. This matches the in-tree debugging workflow since gdbclient.py is
#     normally added to PATH by lunch.
#  2. This ends up calling the actual gdbclient.py script, so the usage
#     message will match the call. Otherwise we would get something like:
#       $$ my_gdb_function -h
#       usage: gdbclient.py ...
gdbclient.py() {
  bdk product gdb -- "$$@"
}

""")

# Inside this template, bdk is re-called to export an updated
# environment.  To do so, the output is packed into an fd and then
# source'd by the shell.  The approach is messier than using <()
# in order to avoid a specific bashism and keep this part dash clean.
# That is yet to be a requirement across all the scripts, though.
ENVSETUP = string.Template("""\
_envsetup_load() {
  # Allow the environment to override the BDK path.
  local bdk="$${BDK_PATH:-${bdk_path}}"
  if ! test -e "$${bdk}/build/core/main.mk"; then
    echo "The BDK cannot be found." 1>&2
    echo "Please supply its path in the BDK_PATH environment variable." 1>&2
    return 1
  fi

  # Make sure we're in the root of a product.
  # If the user explicitly exported a PRODUCT_PATH, then fail if it is wrong.
  if test -n "$${PRODUCT_PATH:-}" && \
     ! test -e "$${PRODUCT_PATH}/AndroidProducts.mk"; then
    echo "Cannot find a product at PRODUCT_PATH: $${PRODUCT_PATH}" 1>&2
    return 1
  fi
  # Otherwise, use the default or check the cwd.
  local product_path="$${PRODUCT_PATH:-${product_path}}"
  if ! test -e "$${product_path}/AndroidProducts.mk"; then
    echo "Checking current directory for a valid product . . ." 1>&2
    product_path=$$PWD
    if ! test -e "$${product_path}/AndroidProducts.mk"; then
      echo "Please source envsetup.sh from the product directory " 1>&2
      echo "or set a valid PRODUCT_PATH in the environment." 1>&2
      return 1
    fi
  fi

  # Find a free file descriptor and use it rather than a tempfile.
  local bdk_tools="$${BDK_TOOLS_PATH:-${tools_path}}"
  fd=$$((`cd /dev/fd && printf -- '%s\n' * | sort -n | tail -1` + 1))
  eval 'exec '$$fd'<<EOFEOF
  $$($${bdk_tools}/cli/bdk product envsetup \\
    --product_path="$${product_path}")
EOFEOF'
  . /dev/fd/$$fd
  eval "exec $$fd<&-"
}
_envsetup_load
""")

SELINUX_DOMAIN = string.Template("""\
# Base domain for a Brillo service.
# You can use '${name}_service' as your service's domain directly,
# or use it as a base for your service's own domain.
type ${name}_service, domain;
type ${name}_service_exec, exec_type, file_type;

# To use '${name}_service' as the domain for your service,
# label the service's executable as '${name}_service_exec' in the 'file_contexts'
# file in this directory.
# brillo_domain() below ensures that executables labelled '${name}_service_exec'
# will be put in the '${name}_service' domain at runtime.

# Allow domain transition from init, and access to D-Bus and Binder.
# See 'te_macros' in $${BDK_PATH}/device/generic/brillo/sepolicy for details.
brillo_domain(${name}_service)

# Allow network access (e.g. opening sockets).
net_domain(${name}_service)

# Allow crash_reporter access to core dump files.
allow_crash_reporter(${name}_service)
""")

SELINUX_FILE_CONTEXTS = string.Template("""\
# Uncomment the line below and replace 'sampleservice' with your service's
# executable.
# /system/bin/sampleservice             u:object_r:${name}_service_exec:s0

# If your service needs to persist data, uncomment the line below and replace
# 'sampleservice' with your service's name or executable.
# /data/misc/sampleservice(/.*)?        u:object_r:${name}_service_data_file:s0
""")
