#
# Copyright (C) 2016 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""This file provides all Target element related parsing."""


import os

from bsp import manifest
from core import user_config
from project import common


# Note: this should be kept in sync with
#   commands.environment.setup.SAMPLE_MAKEFILE_TEMPLATE
PLATFORM_CACHE_FORMAT = os.path.join('{user_configured_platform_cache}',
                                     '{os}.{os_version}',
                                     '{board}.{board_version}')


class Error(common.Error):
    """General Error for targets."""


class BoardError(Error):
    """Raised when there is a problem with the specified board/version."""
    description = 'Invalid board'


class Target(object):
    """A target for a project.

    Specifies what a target consists of, and what platform
    it is intended for.
    """

    # TODO(b/28296932): Rename the 'os' input variable to not conflict with the
    #     import.
    # pylint: disable=redefined-outer-name
    def __init__(self, name, os='', os_version='', board='',
                 board_version=''):
        self._name = name
        self._pack = None
        self._pack_name = ''
        self._os = os
        self._os_version = os_version
        self._board = board
        self._board_version = board_version
        self._build = 'userdebug'
        self._origin = common.Origin()

    def __repr__(self):
        return ('<target name="{}" os="{}" os-version="{}" '
                'board="{}" board-version="{}" build="{}" pack="{}"/>').format(
                    self._name, self._os, self._os_version, self._board,
                    self._board_version, self._build, self._pack_name)

    @property
    def name(self):
        return self._name

    @property
    def pack_name(self):
        return self._pack_name

    @property
    def pack(self):
        return self._pack

    @property
    def origin(self):
        return self._origin

    @property
    def os(self):
        return self._os

    @os.setter
    def os(self, b):
        self._os = b

    @property
    def os_namespace(self):
        return '{}.{}'.format(self.os, self.os_version)

    @property
    def os_version(self):
        return self._os_version

    @os_version.setter
    def os_version(self, b):
        self._os_version = b

    @property
    def board(self):
        return self._board

    @board.setter
    def board(self, b):
        self._board = b

    @property
    def board_version(self):
        return self._board_version

    @board_version.setter
    def board_version(self, b):
        self._board_version = b

    @property
    def board_namespace(self):
        return '{}.{}'.format(self.board, self.board_version)

    @property
    def build_type(self):
        return self._build

    def get_device(self):
        """Get the bsp.Device associated with this Target.

        Returns:
            The bsp.Device associated with this Target.

        Raises:
            BoardError: If the target board and version is not in the manifest.
        """
        bsp_manifest = manifest.Manifest.from_json()
        board = bsp_manifest.devices.get(self.board)
        if not board:
            raise BoardError('unrecognized name "{}". '
                             'Run `bdk bsp list` to see available '
                             'boards.'.format(self.board))
        if board.version != self.board_version:
            raise BoardError('"{}" only has version {} (requested {}).'.format(
                self.board, board.version, self.board_version))
        return board

    def platform_cache(self, *relpath):
        return os.path.join(
            PLATFORM_CACHE_FORMAT.format(
                user_configured_platform_cache=(
                    user_config.USER_CONFIG.platform_cache),
                os=self.os,
                os_version=self.os_version,
                board=self.board,
                board_version=self.board_version),
            *relpath)

    def platform_build_cache(self, *relpath):
        return self.platform_cache(self.build_type, *relpath)

    def load(self, ele):
        """Populates this instance from Element @ele."""
        self._origin = ele.origin.copy()
        ele.limit_attribs(['name', 'pack', 'os', 'board',
                           'os-version', 'board-version', 'build'])
        self._pack_name = ele.get_attrib('pack')
        self._os = ele.get_attrib('os')
        self._os_version = ele.get_attrib('os-version')
        self._board = ele.get_attrib('board')
        self._board_version = ele.get_attrib('board-version')
        if 'build' in ele.attrib:
            self._build = ele.get_attrib('build')

    def create_submap(self, global_packmap):
        """Create a packmap from the global packmap.

        Prior to calling, populate the global packmap with any
        requires OS or Board packs.

        Args:
            global_packmap: A packmap of all available packs.

        Returns:
            A submap of global_packmap for this target, with paths validated.
        """
        aliases = {'os': self.os_namespace, 'board': self.board_namespace}
        packmap = global_packmap.submap(self._pack_name, aliases)
        packmap.check_paths()
        return packmap
